@@ -6,14 +6,14 @@ import random
 import numpy as np
 
 from albumentations.augmentations.keypoints_utils import KeypointsProcessor
-from albumentations.core.serialization import SerializableMeta
+from albumentations.core.serialization import SerializableMeta, get_shortest_class_fullname
 from albumentations.core.six import add_metaclass
 from albumentations.core.transforms_interface import DualTransform
 from albumentations.core.utils import format_args, Params
 from albumentations.augmentations.bbox_utils import BboxProcessor
 from albumentations.core.serialization import SERIALIZABLE_REGISTRY, instantiate_lambda
 
-__all__ = ["Compose", "OneOf", "OneOrOther", "BboxParams", "KeypointParams", "ReplayCompose", "Sequential"]
+__all__ = ["Compose", "SomeOf", "OneOf", "OneOrOther", "BboxParams", "KeypointParams", "ReplayCompose", "Sequential"]
 
 
 REPR_INDENT_STEP = 2
@@ -90,7 +90,7 @@ class BaseCompose:
 
     @classmethod
     def get_class_fullname(cls):
-        return "{cls.__module__}.{cls.__name__}".format(cls=cls)
+        return get_shortest_class_fullname(cls)
 
     def _to_dict(self):
         return {
@@ -118,7 +118,7 @@ class BaseCompose:
 
 
 class Compose(BaseCompose):
-    """Compose transforms and handle all transformations regrading bounding boxes
+    """Compose transforms and handle all transformations regarding bounding boxes
 
     Args:
         transforms (list): list of transformations to compose.
@@ -210,6 +210,22 @@ class Compose(BaseCompose):
         )
         return dictionary
 
+    def get_dict_with_id(self):
+        dictionary = super().get_dict_with_id()
+        bbox_processor = self.processors.get("bboxes")
+        keypoints_processor = self.processors.get("keypoints")
+        dictionary.update(
+            {
+                "bbox_params": bbox_processor.params._to_dict() if bbox_processor else None,  # skipcq: PYL-W0212
+                "keypoint_params": keypoints_processor.params._to_dict()  # skipcq: PYL-W0212
+                if keypoints_processor
+                else None,
+                "additional_targets": self.additional_targets,
+                "params": None,
+            }
+        )
+        return dictionary
+
     def _check_args(self, **kwargs):
         checked_single = ["image", "mask"]
         checked_multi = ["masks"]
@@ -253,6 +269,46 @@ class OneOf(BaseCompose):
         return data
 
 
+class SomeOf(BaseCompose):
+    """Select N transforms to apply. Selected transforms will be called with `force_apply=True`.
+    Transforms probabilities will be normalized to one 1, so in this case transforms probabilities works as weights.
+
+    Args:
+        transforms (list): list of transformations to compose.
+        n (int): number of transforms to apply.
+        replace (bool): Whether the sampled transforms are with or without replacement. Default: True.
+        p (float): probability of applying selected transform. Default: 1.
+    """
+
+    def __init__(self, transforms, n, replace=True, p=1):
+        super(SomeOf, self).__init__(transforms, p)
+        self.n = n
+        self.replace = replace
+        transforms_ps = [t.p for t in transforms]
+        s = sum(transforms_ps)
+        self.transforms_ps = [t / s for t in transforms_ps]
+
+    def __call__(self, force_apply=False, **data):
+        if self.replay_mode:
+            for t in self.transforms:
+                data = t(**data)
+            return data
+
+        if self.transforms_ps and (force_apply or random.random() < self.p):
+            random_state = np.random.RandomState(random.randint(0, 2 ** 32 - 1))
+            transforms = random_state.choice(
+                self.transforms.transforms, size=self.n, replace=self.replace, p=self.transforms_ps
+            )
+            for t in transforms:
+                data = t(force_apply=True, **data)
+        return data
+
+    def _to_dict(self):
+        dictionary = super(SomeOf, self)._to_dict()
+        dictionary.update({"n": self.n, "replace": self.replace})
+        return dictionary
+
+
 class OneOrOther(BaseCompose):
     """Select one or another transform to apply. Selected transform will be called with `force_apply=True`."""
 
